/*
 * Copyright (c) 2008 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "velodyneConfig.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef HAVE_POCOLIBS
#include <portLib.h>
#endif
#include <velodyne/velodyneLib.h>
#include "velodyne_private.h"

/* Calibration data files, built from the XML files provided with each sensor */
#include "velodyneCalibration32.h"
#include "velodyneCalibration64.h"

/* Minimum and maximum distances of the sensors, in meters */
#define VELODYNE_MIN_DISTANCE_32 1.0
#define VELODYNE_MIN_DISTANCE_64 1.0
#define VELODYNE_MAX_DISTANCE_32 70.0
#define VELODYNE_MAX_DISTANCE_64 100.0

/*
 * Precompute the calibration parameters to avoid re-doing it for each point
 * when converting them in Cartesian coordinates
 */

#define VELODYNE_NUM_ROT_ANGLES 36000

int fovFilter32[VELODYNE_NUM_ROT_ANGLES][VELODYNE_NUM_LASERS_32];
int fovFilter64[VELODYNE_NUM_ROT_ANGLES][VELODYNE_NUM_LASERS_64];

double cosHeadingAngleTable32[VELODYNE_NUM_ROT_ANGLES];
double sinHeadingAngleTable32[VELODYNE_NUM_ROT_ANGLES];

double cosVertAngleTable32[VELODYNE_NUM_LASERS_32];
double sinVertAngleTable32[VELODYNE_NUM_LASERS_32];

double cosHeadingAngleTable[VELODYNE_NUM_ROT_ANGLES][VELODYNE_NUM_LASERS_64];
double sinHeadingAngleTable[VELODYNE_NUM_ROT_ANGLES][VELODYNE_NUM_LASERS_64];

const double VELODYNE_OFFSET_SCALE = 0.01;
const double VELODYNE_HEADING_SCALE = 0.01;

double cosVertAngleTable[VELODYNE_NUM_LASERS_64];
double sinVertAngleTable[VELODYNE_NUM_LASERS_64];
double distCorrTable[VELODYNE_NUM_LASERS_64];
double vertOffsetTable[VELODYNE_NUM_LASERS_64];
double horizOffsetTable[VELODYNE_NUM_LASERS_64];

/* Returns the equivalent angle in the [0, 360[ interval */
static double
to_zero_360(double angle) {
	while (angle >= 360.0) angle -= 360.0;
	while (angle < 0.0) angle += 360.0;
	return angle;
}

/* This function is used to sort the laser numbers by decreasing order */
static int
compare_line_indices_32(const void *index1, const void *index2)
{
	int *i1 = (int*) index1;
	int *i2 = (int*) index2;

	/* 1 < 2 => 1 */
	if (vc32[*i1].vertCorrection < vc32[*i2].vertCorrection)
		return 1;
	else
		return -1;
}

/* This function is used to sort the laser numbers by decreasing order */
static int
compare_line_indices_64(const void *index1, const void *index2)
{
	int *i1 = (int*) index1;
	int *i2 = (int*) index2;

	/* 1 < 2 => 1 */
	if (vc64[*i1].vertCorrection < vc64[*i2].vertCorrection)
		return 1;
	else
		return -1;
}

static void
init_calibration_values_32(void)
{
	int i;
	double heading;

	for (i = 0; i < VELODYNE_NUM_ROT_ANGLES; i++) {
		heading = i * VELODYNE_HEADING_SCALE * M_PI / 180.0;
		cosHeadingAngleTable32[i] = cos(heading);
		sinHeadingAngleTable32[i] = sin(heading);
	}
	for (i = 0; i < VELODYNE_NUM_LASERS_32; i++) {
		cosVertAngleTable32[i] = cos(vc32[i].vertCorrection * M_PI / 180.0);
		sinVertAngleTable32[i] = sin(vc32[i].vertCorrection * M_PI / 180.0);
	}
	return;
}

static void
init_calibration_values_64(void)
{
	int i, j;
	double heading;

	for (i = 0; i < VELODYNE_NUM_ROT_ANGLES; i++) {
		for (j = 0; j < VELODYNE_NUM_LASERS_64; j++) {
			heading = (i * VELODYNE_HEADING_SCALE - vc64[j].rotCorrection)
			          * M_PI / 180.0;
			cosHeadingAngleTable[i][j] = cos(heading);
			sinHeadingAngleTable[i][j] = sin(heading);
		}
	}
	for (j = 0; j < VELODYNE_NUM_LASERS_64; j++) {
		cosVertAngleTable[j] = cos(vc64[j].vertCorrection * M_PI / 180.0);
		sinVertAngleTable[j] = sin(vc64[j].vertCorrection * M_PI / 180.0);
		distCorrTable[j] = vc64[j].distCorrection * VELODYNE_OFFSET_SCALE;
		vertOffsetTable[j]
			= vc64[j].vertOffsetCorrection * VELODYNE_OFFSET_SCALE;
		horizOffsetTable[j]
			= vc64[j].horizOffsetCorrection * VELODYNE_OFFSET_SCALE;
	}
	return;
}

/**********************************************************************
 *
 * Pre-computes the various sin and cos that will be used to convert
 * the raw data in Cartesian points.
 * If a sector of data has to be discarded, sets some cos and sin to
 * 0: this allows a cheap test when doing the conversion to discard
 * the points.
 *
 **********************************************************************/

void
velodyne_init_calibration_values(VELODYNE_MODEL model)
{
	switch (model) {
	case VELODYNE_32:
		init_calibration_values_32();
		break;
	case VELODYNE_64:
		init_calibration_values_64();
		break;
	default:
		break;
	}
	return;
}

static void
convert_heading_sector_to_original_reference_frame_32(
		double firstAngle, double lastAngle,
		int * firstAngleOriginal, int * lastAngleOriginal, int * reversedAngles)
{
	*firstAngleOriginal = lround(100 * to_zero_360(-lastAngle - 90.0));
	*lastAngleOriginal = lround(100 * to_zero_360(-firstAngle - 90.0));
	if (*firstAngleOriginal >= *lastAngleOriginal)
		*reversedAngles = 1;
	else
		*reversedAngles = 0;
}

static void
convert_heading_sector_to_original_reference_frame_64(
		double firstAngle, double lastAngle,
		int * firstAngleOriginal, int * lastAngleOriginal, int * reversedAngles)
{
	*firstAngleOriginal = lround(100 * to_zero_360(-lastAngle));
	*lastAngleOriginal = lround(100 * to_zero_360(-firstAngle));
	if (*firstAngleOriginal >= *lastAngleOriginal)
		*reversedAngles = 1;
	else
		*reversedAngles = 0;
}

/*
 * Converts a heading sector to the original reference frame of the Velodyne
 * sensors. The sector is specified by a first and a last angle, in degrees,
 * under the standard assumption that the heading is positive counterclockwise.
 * After the conversion, due to the clockwise sense of the sensors' rotation,
 * the first angle becomes the last one and vice-versa. Variable reversedAngles
 * is set to true if, after this exchange, the first angle is superior to the
 * last one.
 */
void
velodyne_convert_heading_sector_to_original_reference_frame(
		VELODYNE_MODEL model, double firstAngle, double lastAngle,
		int * firstAngleOriginal, int * lastAngleOriginal, int * reversedAngles)
{
	switch (model) {
	case VELODYNE_32:
		convert_heading_sector_to_original_reference_frame_32(
				firstAngle, lastAngle,
				firstAngleOriginal, lastAngleOriginal, reversedAngles);
		break;
	case VELODYNE_64:
		convert_heading_sector_to_original_reference_frame_64(
				firstAngle, lastAngle,
				firstAngleOriginal, lastAngleOriginal, reversedAngles);
		break;
	default:
		break;
	}
	return;
}

static void
init_fov_filters()
{
	memset(fovFilter32, 1,
	       VELODYNE_NUM_ROT_ANGLES * VELODYNE_NUM_LASERS_32 * sizeof(int));
	memset(fovFilter64, 1,
	       VELODYNE_NUM_ROT_ANGLES * VELODYNE_NUM_LASERS_64 * sizeof(int));
	return;
}

static void
add_heading_sector_to_fov_filter_32(double firstDiscardedHeadingAngle,
                                    double lastDiscardedHeadingAngle)
{
	int heading, laserNum;
	int firstAngle, lastAngle;
	int reversedAngles;
	convert_heading_sector_to_original_reference_frame_32(
			firstDiscardedHeadingAngle, lastDiscardedHeadingAngle,
			&firstAngle, &lastAngle, &reversedAngles);

	for (heading = 0; heading < VELODYNE_NUM_ROT_ANGLES; heading++) {
		for (laserNum = 0; laserNum < VELODYNE_NUM_LASERS_32; laserNum++) {
			if (reversedAngles != 0) {
				if (heading > firstAngle || heading < lastAngle)
					fovFilter32[heading][laserNum] = 0;
			} else {
				if (heading > firstAngle && heading < lastAngle)
					fovFilter32[heading][laserNum] = 0;
			}
		}
	}
	return;
}

static void
add_heading_sector_to_fov_filter_64(double firstDiscardedHeadingAngle,
                                    double lastDiscardedHeadingAngle)
{
	int i, j;
	int heading;
	int firstAngle, lastAngle;
	int reversedAngles;
	int rotCorrections[VELODYNE_NUM_LASERS_64];
	convert_heading_sector_to_original_reference_frame_64(
			firstDiscardedHeadingAngle, lastDiscardedHeadingAngle,
			&firstAngle, &lastAngle, &reversedAngles);
	for (i = 0; i < VELODYNE_NUM_LASERS_64; i++)
		rotCorrections[i] = -lround(100 * vc64[i].rotCorrection);
	for (i = 0; i < VELODYNE_NUM_ROT_ANGLES; i++) {
		for (j = 0; j < VELODYNE_NUM_LASERS_64; j++) {
			heading = i + rotCorrections[j];
			while (heading >= VELODYNE_NUM_ROT_ANGLES)
				heading -= VELODYNE_NUM_ROT_ANGLES;
			while (heading < 0)
				heading += VELODYNE_NUM_ROT_ANGLES;
			if (reversedAngles != 0) {
				if (heading > firstAngle || heading < lastAngle)
					fovFilter64[i][j] = 0;
			} else {
				if (heading > firstAngle && heading < lastAngle)
					fovFilter64[i][j] = 0;
			}
		}
	}
	return;
}

/*
 * Adds a heading sector to be discarded by the field-of-view filter.
 */
int
velodyne_add_heading_sector_to_fov_filter(VELODYNE_MODEL model,
                                          double firstDiscardedHeadingAngle,
                                          double lastDiscardedHeadingAngle)
{
	if (firstDiscardedHeadingAngle == lastDiscardedHeadingAngle)
		return VELODYNE_ERR_FOV;
	switch (model) {
	case VELODYNE_32:
		add_heading_sector_to_fov_filter_32(firstDiscardedHeadingAngle,
		                                    lastDiscardedHeadingAngle);
		return VELODYNE_OK;
	case VELODYNE_64:
		add_heading_sector_to_fov_filter_64(firstDiscardedHeadingAngle,
		                                    lastDiscardedHeadingAngle);
		return VELODYNE_OK;
	default:
		return VELODYNE_ERR_MODEL;
	}
}

static int
add_laser_to_fov_filter_32(int laserNum)
{
	int heading;
	if (laserNum < 0 || laserNum >= VELODYNE_NUM_LASERS_32)
		return VELODYNE_ERR_FOV;
	for (heading = 0; heading < VELODYNE_NUM_ROT_ANGLES; heading++)
		fovFilter32[heading][laserNum] = 0;
	return VELODYNE_OK;
}

static int
add_laser_to_fov_filter_64(int laserNum)
{
	int heading;
	if (laserNum < 0 || laserNum >= VELODYNE_NUM_LASERS_64)
		return VELODYNE_ERR_FOV;
	for (heading = 0; heading < VELODYNE_NUM_ROT_ANGLES; heading++)
		fovFilter64[heading][laserNum] = 0;
	return VELODYNE_OK;
}

/*
 * Adds a laser to be discarded by the field-of-view filter.
 */
int
velodyne_add_laser_to_fov_filter(VELODYNE_MODEL model, int laserNum)
{
	switch (model) {
	case VELODYNE_32:
		return add_laser_to_fov_filter_32(laserNum);
	case VELODYNE_64:
		return add_laser_to_fov_filter_64(laserNum);
	default:
		return VELODYNE_ERR_MODEL;
	}
}

static void
get_organization_values_32(int *lineOrder, int *columnOffsets)
{
	int i;
	int lineOrderReverse[VELODYNE_NUM_LASERS_32];
	for (i = 0; i < VELODYNE_NUM_LASERS_32; i++)
		lineOrderReverse[i] = i;
	qsort(lineOrderReverse, VELODYNE_NUM_LASERS_32, sizeof(int),
	      compare_line_indices_32);
	for (i = 0; i < VELODYNE_NUM_LASERS_32; i++) {
		lineOrder[lineOrderReverse[i]] = i;
		columnOffsets[i] = 0;
	}
	return;
}

static void
get_organization_values_64(int *lineOrder, int *columnOffsets)
{
	int i;
	int lineOrderReverse[VELODYNE_NUM_LASERS_64];
	double maxRotCorr;
	double numShotsPerDegree;

	/* lineOrderReverse gives the number of the laser beam ordered by decreasing
	   vertical angle. lineOrder gives for a laser number its line index so that
	   the lines are ordered by decreasing vertical angle */
	for (i = 0; i < VELODYNE_NUM_LASERS_64; i++)
		lineOrderReverse[i] = i;
	qsort(lineOrderReverse, VELODYNE_NUM_LASERS_64, sizeof(int),
	      compare_line_indices_64);
	for (i = 0; i < VELODYNE_NUM_LASERS_64; i++)
		lineOrder[lineOrderReverse[i]] = i;

	/* columnOffsets gives the offset to apply to the columns indices so that a
	   column corresponds to the same azimuth angle */
	maxRotCorr = vc64[0].rotCorrection;
	for (i = 1; i < VELODYNE_NUM_LASERS_64; ++i)
		if (vc64[i].rotCorrection > maxRotCorr)
			maxRotCorr = vc64[i].rotCorrection;
	numShotsPerDegree = VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64
	                    * VELODYNE_MEAN_NUM_PACKETS_PER_REVOLUTION_64 / 360;
	/* 6 is the number of laser shots per packet, 730 is the mean number of
	   packets acquired during a full scan when the sensors spins at 5 Hz */
	for (i = 0; i < VELODYNE_NUM_LASERS_64; i++)
		columnOffsets[lineOrder[i]]
				= lround((maxRotCorr - vc64[i].rotCorrection)
				         * numShotsPerDegree);
	return;
}

/* columnOffsets is the VELODYNE_NUM_LASERS_64 array of offsets to apply to
   the column indices so that a column correspond to the same azimuth
   angle. maxOffset is the maximum of these offsets */
void
velodyne_get_organization_values(VELODYNE_MODEL model, int *lineOrder,
                                 int *columnOffsets)
{
	switch (model) {
	case VELODYNE_32:
		get_organization_values_32(lineOrder, columnOffsets);
		break;
	case VELODYNE_64:
		get_organization_values_64(lineOrder, columnOffsets);
		break;
	default:
		break;
	}
	return;
}

/*
 * Function to convert raw velodyne data into cartesian coordinates in the
 * sensor frame. See the Velodyne documentation or reference [1] for explanation
 * of the way the conversion is achieved.
 *
 * [1]: N. Muhammad and S. Lacroix. "Calibration of a rotating multi-beam
 * Lidar", IEEE/RSJ International Conference on Intelligent Robots and Systems,
 * 2010. 
 */

const double VELODYNE_DIST_SCALE = 0.002;

static inline void
compute_cartesian_point_32(int laserNum, u_int16_t heading, u_int16_t dist,
                           u_int8_t intensity, VELODYNE_CARTESIAN_POINT *point)
{
	double distance = dist * VELODYNE_DIST_SCALE;

	if (dist == 0
	    || distance < VELODYNE_MIN_DISTANCE_32
	    || distance > VELODYNE_MAX_DISTANCE_32
	    || fovFilter32[heading][laserNum] == 0) {
		point->status = VELODYNE_BAD_POINT;
		return;
	}
	double cosVertAngle = cosVertAngleTable32[laserNum];
	double sinVertAngle = sinVertAngleTable32[laserNum];
	double cosHeadingAngle = cosHeadingAngleTable32[heading];
	double sinHeadingAngle = sinHeadingAngleTable32[heading];
	double xyDistance = distance * cosVertAngle;
	point->x = (float) -(xyDistance * sinHeadingAngle);
	point->y = (float) -(xyDistance * cosHeadingAngle);
	point->z = (float) (distance * sinVertAngle);
	point->intensity = intensity;
	point->status = VELODYNE_GOOD_POINT;
	return;
}

static inline void
compute_cartesian_point_64(int laserNum, u_int16_t heading, u_int16_t dist,
                           u_int8_t intensity, VELODYNE_CARTESIAN_POINT *point)
{
	double distance = dist * VELODYNE_DIST_SCALE + distCorrTable[laserNum];

	if (dist == 0
	    || distance < VELODYNE_MIN_DISTANCE_64
	    || distance > VELODYNE_MAX_DISTANCE_64
	    || fovFilter64[heading][laserNum] == 0) {
		point->status = VELODYNE_BAD_POINT;
		return;
	}

	double cosVertAngle = cosVertAngleTable[laserNum];
	double sinVertAngle = sinVertAngleTable[laserNum];
	double vertOffset = vertOffsetTable[laserNum];
	double horizOffset = horizOffsetTable[laserNum];
	double cosHeadingAngle = cosHeadingAngleTable[heading][laserNum];
	double sinHeadingAngle = sinHeadingAngleTable[heading][laserNum];

	double xyDistance = distance * cosVertAngle - vertOffset * sinVertAngle;

	point->status = VELODYNE_GOOD_POINT;
	/* Warning: The two lines above take into account the fact that the original
	   sensor frame is as follows: the Y axis is oriented opposite to the
	   wire connector, the heading is 0 along this Y axis, and the heading is
	   positive clock-wise (strange convention fron Velodyne). Doing x = y, and
	   y = -x defines a usual sensor frame, with the X axis heading opposite to
	   the wire connector */
	point->x = (float) (xyDistance * cosHeadingAngle + horizOffset * sinHeadingAngle);
	point->y = (float) -(xyDistance * sinHeadingAngle - horizOffset * cosHeadingAngle);
	point->z = (float) (distance * sinVertAngle + vertOffset * cosVertAngle);
	point->intensity = intensity;
	return;
}

/*----------------------------------------------------------------------*/

/**
 * \brief Get access to the lidar device using a pcap file.
 */
VELODYNE_INTERFACE *
velodyne_open_file(VELODYNE_MODEL model, const char *name)
{
	VELODYNE_INTERFACE *dev;

	dev = (VELODYNE_INTERFACE *)malloc(sizeof(VELODYNE_INTERFACE));
	if (dev == NULL) {
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_NO_MEMORY);
#endif
		return NULL;
	}
	if (velodyne_pcap_file_open(name, &dev->pcap) == -1) {
		free(dev);
		return NULL;
	}
	dev->type = VELODYNE_FILE;
	dev->model = model;
	velodyne_init_calibration_values(dev->model);
	init_fov_filters();
	return dev;
}

/**
 * \brief Get access to the lidar device using the network.
 */
VELODYNE_INTERFACE *
velodyne_open_network(VELODYNE_MODEL model)
{
	VELODYNE_INTERFACE *dev;

	dev = (VELODYNE_INTERFACE *)malloc(sizeof(VELODYNE_INTERFACE));
	if (dev == NULL) {
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_NO_MEMORY);
#endif
		return NULL;
	}
	dev->fd = velodyne_setup_socket();
	if (dev->fd == -1) {
		free(dev);
		return NULL;
	}
	dev->type = VELODYNE_NETWORK;
	dev->model = model;
	velodyne_init_calibration_values(dev->model);
	init_fov_filters();
	return dev;
}

/*----------------------------------------------------------------------*/

/**
 * \brief Terminate the access to the device
 */
void
velodyne_close(VELODYNE_INTERFACE *device)
{
	switch (device->type) {
	case VELODYNE_FILE:
		velodyne_close_pcap(device->pcap);
		break;
	case VELODYNE_NETWORK:
		velodyne_close_socket(device->fd);
		break;
	default:
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
		break;
	}
	free(device);
	return;
}

/*----------------------------------------------------------------------*/
/**
 * \brief Read a packet from the device 
 */
int
velodyne_read_packet(const VELODYNE_INTERFACE *device, VELODYNE_PACKET *p)
{
	switch (device->type) {
	case VELODYNE_FILE:
		return velodyne_pcap_read_packet(device->pcap, p);
	case VELODYNE_NETWORK:
		return velodyne_udp_rcv(device->fd, p, 0);
	default:
#ifdef HAVE_POCOLIBS
		errnoSet(S_velodyneLib_INVALID_DEVICE);
#endif
		return -1;
	}
	return 0;
}

/*----------------------------------------------------------------------*/
/*
 * 
 * A packet of data contains VELODYNE_NUM_FIRINGS_PER_PACKET blocks of VELODYNE_NUM_LASERS_PER_FIRING
 * points (i.e. 12 x 32 = 384 points). Each block comes either for the upper or
 * the lower sets of lasers (the sensor has VELODYNE_NUM_LASERS_64 = 64
 * lasers), and is encoded in the structure VELODYNE_FIRING, that contains:
 * - whether it is an upper or lower block 
 * - its rotational position, i.e. the heading angle of the rotating sensor when
 * data are acquired 
 * - VELODYNE_NUM_LASERS_PER_FIRING structures VELODYNE_LASER
 * 
 */

void
velodyne_print_packet(const VELODYNE_PACKET *p)
{
	int i, j;
	const VELODYNE_RAW_PACKET *packet = &p->data;
	struct tm localTime;
	char buf[100];

	localtime_r(&p->ts.tv_sec, &localTime);
	strftime(buf, sizeof(buf), "%F %T", &localTime);
	printf("%s.%06lu\n", buf, (unsigned long)p->ts.tv_usec);
	for (i = 0; i < VELODYNE_NUM_FIRINGS_PER_PACKET; i++) {
		printf("Block %d, %s, rotational_pos: %d:\n", i,
		       packet->firings[i].block_id == VELODYNE_UPPER ?
		       "upper" :
		       packet->firings[i].block_id == VELODYNE_LOWER ?
		       "lower" : "unknown",
		       packet->firings[i].rotational_pos);
		printf("Distances:\t");
		for (j = 0; j < VELODYNE_NUM_LASERS_PER_FIRING; j++)
			printf("%d\t", packet->firings[i].lasers[j].distance);
		printf("\nIntensities:\t");
		for (j = 0; j < VELODYNE_NUM_LASERS_PER_FIRING; j++)
			printf("%d\t", packet->firings[i].lasers[j].intensity);
		printf("\n");
	}
	printf("Status: %#.2x, %#.2x, %#.2x, %#.2x, %#.2x, %#.2x\n",
	       packet->status[0], packet->status[1], packet->status[2],
	       packet->status[3], packet->status[4], packet->status[5]);
	printf("\n\n");
	return;
}

void
velodyne_print_cartesian_packet(const VELODYNE_CARTESIAN_PACKET *packet)
{
	const VELODYNE_CARTESIAN_POINT *p;
	int i;

	for (i = 0; i < VELODYNE_NUM_LASERS_PER_PACKET; i++) {
		p = &(packet->points[i]);
		if (p->status == VELODYNE_GOOD_POINT)
			printf("%.3f\t%.3f\t%.3f\t%d\n", p->x, p->y, p->z, p->intensity);
	}
	printf("\n\n");
	return;
}

/*----------------------------------------------------------------------*/
/*
 * Converts a VELODYNE_PACKET into a VELODYNE_CARTESIAN_PACKET. Points
 * are expressed in the sensor frame.
 */

void
velodyne_to_cartesian_packet(VELODYNE_MODEL model, const VELODYNE_PACKET *pkt,
                             VELODYNE_CARTESIAN_PACKET *cartesianPkt)
{
	const VELODYNE_RAW_PACKET *packet = &pkt->data;
	VELODYNE_CARTESIAN_POINT *p = cartesianPkt->points;
	const VELODYNE_FIRING *s;
	int i, j, n;

	switch (model) {
	case VELODYNE_32:
		for (i = 0; i < VELODYNE_NUM_FIRINGS_PER_PACKET; i++) {
			s = &(packet->firings[i]);
			for (j = 0; j < VELODYNE_NUM_LASERS_PER_FIRING; j++) {
				compute_cartesian_point_32(j, s->rotational_pos,
				                           s->lasers[j].distance,
				                           s->lasers[j].intensity, p);
				p++;
			}
		}
		break;
	case VELODYNE_64:
		for (i = 0; i < VELODYNE_NUM_FIRINGS_PER_PACKET; i++) {
			s = &(packet->firings[i]);
			for (j = 0; j < VELODYNE_NUM_LASERS_PER_FIRING; j++) {
				n = j + (s->block_id == VELODYNE_UPPER ? 0 : 1)
				        * VELODYNE_NUM_LASERS_PER_FIRING;
				compute_cartesian_point_64(n, s->rotational_pos,
				                           s->lasers[j].distance,
				                           s->lasers[j].intensity, p);
				p++;
			}
		}
		break;
	default:
		break;
	}
	cartesianPkt->ts = pkt->ts;
	return;
}
